import os # rozne systemove rozhrania
import time # praca s casom
from threading import Thread # paralelizmus zalozeny na vlaknach

import dash # Dash framework
import dash_html_components as html
import dash_core_components as dcc
from dash.dependencies import Input,Output

# Scapy - send, sniff, dissect, forge network packets
from scapy.all import *

import pandas as pd # Pandas
import numpy as np # NumPy

# Plotly
import plotly.graph_objects as go
import plotly.express as px

# inicializacia Pandas DataFrame
net_data = pd.DataFrame(columns = ["netid","BSSID","Manf","SSID","dBm_Signal","Channel","Crypto"])
# index = netid
net_data.set_index("netid",inplace = True)
# zoznam aktualne najdenych sieti
act_ssid = []

# inicializacia Dash aplikacie
app = dash.Dash(__name__)
app.config.suppress_callback_exceptions = True

# callable object - zmena WiFi kanalu
def my_hopper(iface):
    global act_ssid
    ch = 1
    while True:
        # vykonanie systemoveho prikazu
        os.system('iwconfig %s channel %d' % (iface,ch))
        #print('aktualny kanal %d' % (ch))
        ch = ch + 1
        if ch > 13:
            ch = 1
            # vymazanie siete, ktora uz nie je v dosahu
            for i in net_data.SSID: # pre vsetky polozky v DataFrame.SSID
                if i not in act_ssid: # ak polozka nie je v zozname aktualne najdenych sieti
                    net_data.drop(i,inplace = True) # vymazanie riadku z DataFrame
                    #print('mazem '+i)
            act_ssid.clear() # vymazanie zoznamu aktualne najdenych sieti
        time.sleep(0.5) # 0.5s pauza

# vyrobca vysielaca
def get_manf(bss):
    file = open('data/bssid.txt','r')
    for line in file.read().splitlines():
        if line.split('~')[0][7] == bss[7]:
            return line.split('~')[1].rstrip('\n')
    return 'n/a'

# ziskanie informacii z prijatych paketov
def packet_handler(pkt):
    global act_ssid
    #print('packet')
    if pkt.haslayer(Dot11Beacon):
        # MAC adresa vysielaca
        bssid = pkt[Dot11].addr2
        # vyrobca vysielaca
        manf = get_manf(bssid.upper())
        # SSID = ID siete
        ssid = pkt[Dot11Elt].info.decode()
        # netid = ssid siete
        netid = ssid
        # vysielaci vykon
        try:
            dbm_signal = pkt.dBm_AntSignal
        except:
            dbm_signal = 'n/a'
        # vysielaci kanal a pouzite sifrovanie
        try:
            stats = pkt[Dot11Beacon].network_stats()
            #print(stats)
            # vysielaci kanal
            channel = stats.get('channel')
            # pouzite sifrovanie
            cr = stats.get('crypto')
            np.warnings.filterwarnings('ignore',category = np.VisibleDeprecationWarning)
            crypto = np.array(list(cr),dtype = object)
        except:
            channel = 'n/a'
            crypto = 'n/a'
        #print(netid,bssid,ssid,dbm_signal,channel,cr)
        net_data.loc[netid] = (bssid,manf,ssid,dbm_signal,channel,crypto)
        # doplnenie siete do zoznamu aktualne najdenych sieti
        if ssid not in act_ssid:
            act_ssid.append(ssid)

# callable object - ziskanie informacii z prijatych paketov
def my_sniffer(iface):
    # iface = interface
    # prn = funkcia aplikovana na kazdy paket
    sniff(iface = iface,prn = packet_handler)

# definicia Dash layoutu
app.layout = html.Div(
    # 1 riadok
    children = [html.Div(className = 'row',
                         children = [# 1. stlpec
                                     html.Div(className = 'six columns',style = { 'margin' : 10 },
                                              children = [# 1 riadok
                                                          html.H3('WiFi skener'),
                                                          dcc.Graph(id = 'live-update-table')
                                              ]
                                     ),
                                     # 2. stlpec
                                     html.Div(className = 'five columns',
                                              children = [# 1 riadok
                                                          dcc.Graph(id = 'live-update-graph'),
                                                          dcc.Interval(
                                                              id = 'interval-table',
                                                              interval = 5000, # updatuje n_intervals kazdych 5s
                                                              n_intervals = 0
                                                          )
                                              ]
                                     )
                         ]
                )
    ]
)

# Callback funkcia pre tabulku
@app.callback(Output('live-update-table','figure'),[Input('interval-table','n_intervals')])
def update_table_object(n):
    # tabulka
    fig = go.Figure(data=[go.Table(header = dict(values = ['MAC','Výrobca','ID siete','Výkon','Kanál','Šifrovanie']),
                                   cells = dict(values = [net_data.BSSID,
                                                          net_data.Manf,
                                                          net_data.SSID,
                                                          net_data.dBm_Signal,
                                                          net_data.Channel,
                                                          net_data.Crypto]),
                                   columnwidth = [120,180,180,50,50,120])
                         ]
                   )
    # nastavenie stylu tabulky
    fig.update_layout(template = 'plotly_dark',
                      paper_bgcolor = 'rgba(0,0,0,100)',
                      margin = {'l' : 10,'r' : 10,'t' : 10,'b' : 10},
                      showlegend = False)
    # aktualizacia dcc.Graph(figure)
    return fig

# Callback funkcia pre graf
@app.callback(Output('live-update-graph','figure'),[Input('interval-table','n_intervals')])
def update_graph_object(n):
    # stlpcovy graf
    fig = go.Figure(data=[go.Bar(x = net_data.SSID,
                                 y = net_data.dBm_Signal,
                                 marker_color = net_data.dBm_Signal,
                                 marker_colorscale = px.colors.diverging.RdYlGn)
                         ]
                   )
    # nastavenie stylu grafu
    fig.update_layout(template = 'plotly_dark',
                      paper_bgcolor = 'rgba(0,0,0,0)',
                      plot_bgcolor = 'rgba(0,0,0,0)',
                      title = {'text' : '','font' : {'color' : 'white'},'x' : 0.5},
                      margin = {'l' : 10,'r' : 10,'t' : 75,'b' : 10},
                      xaxis_title = 'Identifikátor siete',
                      yaxis_title = 'Vysielací výkon (dBm)')
    # aktualizacia dcc.Graph(figure)
    return fig

if __name__ == "__main__":
    # wifi interface
    iface = "wlp2s0"
    # zmena WiFi kanalu v samostatnom vlakne
    hopper = Thread(target = my_hopper,args = (iface,))
    hopper.daemon = True
    hopper.start()
    # ziskanie informacii z prijatych paketov v samostatnom vlakne
    sniffer = Thread(target = my_sniffer,args = (iface,))
    sniffer.daemon = True
    sniffer.start()

    # spustenie Dash aplikacie
    app.run_server(debug = True)